/*The code was written and debugged by John C. Martin at the University of
  Minnesota (copyright 2004,2005).  You are free to distribute this source
  as long as this notice remains in tact.  End users of this code agree to
  take responsibility for how this code is used and John C. Martin and the
  University of Minnesota bear no liability.*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "mkvector.h"
#include "fitsio.h"

int locateghost(double **pix,long naxes[3],
		double hc, double hr,double pa,double d,
		double *gc,double *gr,double *sgc, double *sgr,double *ap);
void PeakUp(double **apix, long *naxes,double *x,double *y, float frad);
int readfits(char *fn,int naxis,long naxes[3],double **array,int hdu,
             int *status);
int getimgdim(char *fn,int *naxis,long naxes[3],int hdu,int *status);
int writefits(char *fn, double **ipix,long naxes[3],int *status);
int rmghost(double **pix,int flag,double amp,double r0,
            double c0,double sgc,double sgr);
float SumApp(double **apix, long *naxes, float x0,float y0,int filtype,
	     float a, float b,int psum,int *nl);
int writeoutfits(char *fn, char *fc, double **ipix,
                 long naxes[3],int *status);

int main(int argc, char *argv[])
{
  char *fln,*fout,*cmd;
  int status=0;
  int naxis,hdu=2;
  long naxes[3];
  double **pix,**mask;
  double hr=0,hc=0,gr=0,gc=0,sgr=0,sgc=0,ap=0;
  float amp=0,f1,f2;
  int i;
  double pa=(-35.7/180.)*3.1415269; //position angle of the ghost
  double d=105.; //separation of ghost (reduced pixels)
  /*double cwin=52;
  double rwin=30;
  double cut[200];*/
  int findit=1;
  int debug=0;
  int modelonly=0;

  fln=(char*)malloc(sizeof(char)*80);
  fout=(char*)malloc(sizeof(char)*80);
  cmd=(char*)malloc(sizeof(char)*80);
  if (argc<=1) {
    printf("usage:  ghostbegone <input file> [-g] [-m] [-f] [-c c,r] [-s sc,sr]\n");
    printf("                    [-p amplitude] [-o output file]\n\n");
    printf("Locate and remove the H-alpha ghost as modeled by a 2-D gaussian from\n a STIS CCD spectrum with the following options:\n\n");
    printf("  -c c,r    Force the ghost center to be at column c, row r\n");
    printf("            Default is to find ghost based on position of H-alpha.\n");
    printf("            Invoking this option overrides the default.\n");
    printf("  -f        Force ghostbgone to use the parameters entered for the \n");
    printf("            ghost.  Default is to either derive the parameters from \n");
    printf("            scratch or use the entered parameters as an initial guess.\n");
    printf("  -g        Run in debug mode (extra output)\n");
    printf("  -m        The output file should only be a single FITS array containig\n");
    printf("            the model itself.  If this is not invoked then the output\n");
    printf("            FITS file is a replicate of the input file with the ghost\n");
    printf("            model subtracted from the flux array.\n");
    printf("  -o file   Explicitly set the name of the output file.  Default = ghost.fits\n");
    printf("  -p value  Explicitly set the peak aplitude of the ghost.  If value is\n");
    printf("            not zero then it overrides any other determination\n");
    printf("            regaurdless of -f or not\n");
    printf("  -s sc,sr  Explicitly set the values of the sigmas (sc = collum-wise\n");
    printf("            and sr = row-wise) in the 2D gaussian ghost model.\n");
    printf("\nNote that files which have been run through ghostbgone include the\n  FITS history entry: \"H-alpha ghost removed by ghostbgone from <original \n  filename>.\"\n");
    return(0);
  }
  strcpy(fout,"!ghost.fits");
  strcpy(fln,argv[1]);
  i = argc - 2;
  while (i>=1) {
    strcpy(cmd,argv[argc-i]);
    i--;
    if (cmd[0]!='-') printf("Unknown:  Skipping [%s]\n",cmd);
    else {
      if (cmd[1]=='o') {
        if (cmd[2]==(char)NULL) {strcpy(fout,argv[argc-i]);i--;}
        else strcpy(fout,cmd+2);
        if (debug) printf("Set fout = %s\n",fout);
      }
      if (cmd[1]=='c') {
        if (cmd[2]==(char)NULL) {strcpy(cmd,argv[argc-i]);i--;}
        else cmd=cmd+2;
        sscanf(cmd,"%f,%f",&f1,&f2);
	gc=f1;gr=f2;
        if (debug) printf("Set ghost center = %g,%g\n",gc,gr);
      }
      if (cmd[1]=='p') {
        if (cmd[2]==(char)NULL) {strcpy(cmd,argv[argc-i]);i--;}
        else cmd=cmd+2;
        amp=atof(cmd);
        if (debug) printf("Set amp = %g\n",amp);
      }
      if (cmd[1]=='s') {
        if (cmd[2]==(char)NULL) {strcpy(cmd,argv[argc-i]);i--;}
        else cmd=cmd+2;
        sscanf(cmd,"%f,%f",&f1,&f2);
	sgc=f1;sgr=f2;
        if (debug) printf("Set ghost sigmas = %g,%g\n",sgc,sgr);
      }
      if (cmd[1]=='g') {debug=1;if (debug) printf("Debug ON\n");}
      if (cmd[1]=='f') {findit=0; if (debug) printf("FindIt OFF\n");}     
      if (cmd[1]=='m') {modelonly=1;if (debug) printf("ModelOnly ON\n");}      
    }
  }
  if (fout[0]!='!')
    if (fopen(fout,"r")!=NULL) {
      fprintf(stderr,"ERROR:  File %s exists and overwrite not set\n",fout);
      return(-1);
    }

  /*Get the input image*/
  getimgdim(fln,&naxis,naxes,hdu,&status);
  pix=dvector2d(0,naxes[1],0,naxes[0]);
  mask=dvector2d(0,naxes[1],0,naxes[0]);
  readfits(fln,naxis,naxes,pix,hdu,&status);
  if (status) {
    ffrprt(stderr, status); /* print any error message */
    return(-2);
  } 

  if (findit) {
    if (debug) printf("Fitting to find location of H-alpha ghost\n");
    /*Locate Halpha if not specified*/
    if (hc==0) hc=334.;
    if (hr==0) hr=512.;
    PeakUp(pix,naxes,&hc,&hr,5.);
    if (debug) printf("Ha center: %g,%g\n",hc,hr);
    /*Find ghost parameters*/
    locateghost(pix,naxes,hc,hr,pa,d,&gc,&gr,&sgc,&sgr,&ap);
    if (debug) printf("[%g %g %g %g %g]\n",gc,gr,sgc,sgr,ap);
    if (amp==0) {
      amp=1.5*ap;
      if (debug) printf("Setting amp = %g from locateghost fit\n",amp);
    }
  } else if ((gc==0)&&(sgc==0)) {
    printf("Error:  If you don't want me to find the ghost then you need\n");
    printf("        to give me the coordinates and sigmas of the ghost with\n");
    printf("        the -c and -s options.\n");
    return(-1);
  }
  if (amp==0) {
    amp=0.66*SumApp(pix,naxes,(float)gc,(float)gr,1,8.,0,1,&i);
    if (debug) printf("Setting amp=%g from SumApp on center\n",amp);
  }
  printf("Ghost:\n  Center = (%g,%g)\n  Width: (%g,%g)\n  Peak = %e\n"
	 ,gc,gr,sgc,sgr,amp); 
  rmghost(mask,1,(double)amp,gr,gc,sgc,sgr);
  rmghost(pix,0,(double)amp,gr,gc,sgc,sgr);
  if (modelonly) {
    writefits(fout,mask,naxes,&status);
  } else {
    writeoutfits(fout,fln,pix,naxes,&status);
  }
  return(0);
}
